#!/usr/bin/env python3

from pathlib import Path
from typing import List, Union

from . import util
from .config import Config
from .mod import EXPANSION_PACKAGES, Mod, ModFolder
from .modsconfig import ModsConfig
from .steam import SteamDownloader, WorkshopResult


class Manager:
    def __init__(self, config: Config):
        if not isinstance(config, Config):
            raise Exception("Must pass Config object to Manager")
        self.config = config
        if self.config.modsconfig_path:
            self.modsconfig = ModsConfig(self.config.modsconfig_path)

    def install_mod(self, steam_cache: Path, steamid: int):
        if not steamid:
            raise Exception("Missing SteamID")
        mod = Mod.create_from_path(steam_cache / str(steamid))

        dest_path = None
        if self.config.USE_HUMAN_NAMES and mod and mod.packageid:
            dest_path = self.config.mod_path / mod.packageid
        else:
            dest_path = self.config.mod_path / str(steamid)

        if dest_path:
            util.copy(
                steam_cache / str(steamid),
                dest_path,
                recursive=True,
            )
        else:
            print(f"Unable to install mod: {steamid}")
            return False
        return True

    def remove_mod(self, mod: Mod):
        if not self.config.mod_path:
            raise Exception("Game path not defined")

        installed_mods = ModFolder.read(self.config.mod_path)
        removal_queue = [n for n in installed_mods if n == mod]

        print("\n")
        for m in removal_queue:
            print(f"Uninstalling {mod.title()}")
            mod_absolute_path = self.config.mod_path / m.dirname
            if mod_absolute_path:
                util.remove(mod_absolute_path)

            steamid_path = self.config.mod_path / str(m.steamid)
            if m.steamid and steamid_path.exists():
                util.remove(self.config.mod_path / str(m.steamid))

            pid_path = self.config.mod_path / m.packageid
            if self.config.USE_HUMAN_NAMES and m.packageid and pid_path.exists():
                util.remove(pid_path)

    def remove_mods(self, queue: List[Mod]):
        for mod in queue:
            if isinstance(mod, WorkshopResult):
                mod = Mod.create_from_workshorp_result(mod)
            self.remove_mod(mod)

    def sync_mods(self, queue: Union[List[Mod], List[WorkshopResult]]):
        steam_mods, steam_cache_path = SteamDownloader.download(
            [mod.steamid for mod in queue if mod.steamid]
        )

        for mod in queue:
            if isinstance(mod, WorkshopResult):
                new_mod = [m for m in steam_mods if m.steamid == mod.steamid]
                if len(new_mod) == 1:
                    mod = new_mod[0]
                else:
                    mod = Mod(steamid=mod.steamid)
            if not isinstance(mod.steamid, int):
                continue
            success = False
            try:
                self.remove_mod(mod)
                success = self.install_mod(steam_cache_path, mod.steamid)
            except FileNotFoundError:
                print(
                    f"Unable to download and install {mod.title()}\n\tDoes this mod still exist?"
                )
            if success:
                print(f"Installed {mod.title()}")

    def _mod_config_state(self, mods):
        return [m for _, m in self._mod_config_state_dict(mods).items()]

    def _mod_config_state_dict(self, mods):
        if self.modsconfig:
            enabled_mods = self._enabled_mod_pids()
            for k, v in mods.items():
                if k in enabled_mods:
                    v.enabled = True
                else:
                    v.enabled = False
        return mods

    def installed_mods(self):
        mods = ModFolder.read_dict(self.config.mod_path)
        return self._mod_config_state(mods)

    def installed_mods_dict(self):
        mods = ModFolder.read_dict(self.config.mod_path)
        return self._mod_config_state_dict(mods)

    def search_installed(self, term):
        mods = ModFolder.search_dict(self.config.mod_path, term)
        return self._mod_config_state(mods)

    def _enabled_mod_pids(self):
        return [k for k in self.modsconfig.mods]

    def enabled_mods(self):
        installed_mods = self.installed_mods_dict()
        l = list()
        for n in self.modsconfig.mods:
            try:
                l.append(installed_mods[n])
            except KeyError:
                continue
        return l

    def disabled_mods(self):
        enabled_mods = self._enabled_mod_pids()
        installed_mods = self.installed_mods()
        return util.list_loop_exclusion(installed_mods, enabled_mods)

    def _enable_mod(self, mod: Union[str, Mod]):
        if isinstance(mod, str):
            mod = Mod(packageid=mod)
        if not mod.packageid:
            raise Exception("No package id for specifed mod")
        self.modsconfig.enable_mod(mod)

    def enable_mods(self, mods):
        for n in mods:
            print("Enabling " + n.title())
            self._enable_mod(n)
        print("Updating ModsConfig.xml")
        self.modsconfig.write()

    def _disable_mod(self, mod: Union[str, Mod]):
        if isinstance(mod, str):
            mod = Mod(packageid=mod)
        if not mod.packageid:
            raise Exception("No package id for specifed mod")
        self.modsconfig.disable_mod(mod)

    def disable_mods(self, mods):
        for n in mods:
            print("Disabling " + n.title())
            self._disable_mod(n)
        print("Updating ModsConfig.xml")
        self.modsconfig.write()

    def verify_mods(self):
        return self.modsconfig.verify_state(self.installed_mods())

    def sort_mods(self):
        self.modsconfig.autosort(self.installed_mods(), self.config)

    def order_all_mods(self):
        installed_mods = self.installed_mods()
        enabled_mods = self._enabled_mod_pids()
        sorted_mods = self._order_mods(enabled_mods, installed_mods)
        for m in installed_mods:
            if m not in sorted_mods:
                sorted_mods.append(m)
        return sorted_mods

    def _order_mods(self, enabled_mods, installed_mods):
        sorted_mods = []
        for m in EXPANSION_PACKAGES:
            m.enabled = True
        for m in enabled_mods:
            for im in installed_mods + EXPANSION_PACKAGES:
                if m == im:
                    sorted_mods.append(im)
                    break
        return sorted_mods

    def order_mods(self):
        enabled_mods = self._enabled_mod_pids()
        installed_mods = self.installed_mods()
        return self._order_mods(enabled_mods, installed_mods)



#!/usr/bin/env python3

from __future__ import annotations

import csv
import re
from abc import ABC, abstractmethod
from collections.abc import MutableSequence
from pathlib import Path
from typing import Any, Generator, Iterator, cast

from .mod import Mod


class ModListSerializer(ABC):
    @classmethod
    @abstractmethod
    def parse(cls, text: str) -> Generator[Mod, None, None]:
        pass

    @classmethod
    @abstractmethod
    def serialize(cls, mods: MutableSequence) -> Generator[str, None, None]:
        pass


class CsvStringBuilder:
    def __init__(self):
        self.value = []

    def write(self, row: str) -> None:
        self.value.append(row)

    def pop(self) -> None:
        return self.value.pop()

    def __iter__(self) -> Iterator[str]:
        return iter(self.value)


class ModListV2Format(ModListSerializer):
    HEADER = {"PACKAGE_ID": 0, "STEAM_ID": 1, "REPO_URL": 2}
    MAGIC_FLAG = "RMM_V2_MODLIST"

    @classmethod
    def parse(cls, text: str) -> Generator[Mod, None, None]:
        reader = csv.reader(text.split("\n"))
        for parsed in reader:
            try:
                yield Mod(
                    packageid=parsed[cls.HEADER["PACKAGE_ID"]],
                    steamid=int(parsed[cls.HEADER["STEAM_ID"]]),
                    repo_url=parsed[cls.HEADER["REPO_URL"]] if not "" else None,
                )
            except IndexError:
                if parsed:
                    print("Unable to import: ", parsed)
                continue
            except ValueError:
                yield Mod(
                    packageid=parsed[cls.HEADER["PACKAGE_ID"]],
                    repo_url=parsed[cls.HEADER["REPO_URL"]] if not "" else None,
                )
                continue

    @classmethod
    def serialize(cls, mods: MutableSequence) -> Generator[str, None, None]:
        buffer = CsvStringBuilder()
        writer = csv.writer(cast(Any, buffer))
        for m in mods:
            writer.writerow(cls.format(m))
            yield buffer.pop().strip()

    @classmethod
    def format(cls, mod: Mod) -> list[str]:
        return cast(
            list[str],
            [
                mod.packageid,
                str(mod.steamid) if not None else "",
                mod.repo_url if not None else "",
            ],
        )


class ModListV1Format(ModListSerializer):
    STEAM_ID = 0

    @classmethod
    def parse(cls, text: str) -> Generator[Mod, None, None]:
        name_exp = re.compile("(.*) by (.*)")
        for line in text.split("\n"):
            parsed = line.split("#", 1)
            name = None
            author = None
            if len(parsed) == 2:
                matches = re.findall(name_exp, parsed[1])
                if matches and len(matches[0]) == 2:
                    name = matches[0][0]
                    author = matches[0][1]
            try:
                yield Mod(
                    steamid=int(
                        parsed[cls.STEAM_ID]
                        .strip()
                        .encode("ascii", errors="ignore")
                        .decode()
                    ),
                    name=name,
                    author=author,
                )
            except ValueError:
                if line:
                    print("Unable to import: ", line)
                continue

    @classmethod
    def serialize(cls, mods: MutableSequence) -> Generator[str, None, None]:
        for m in mods:
            yield cls.format(m)

    @classmethod
    def format(cls, mod: Mod) -> str:
        return "{}# {} by {} ".format(str(mod.steamid), mod.name, mod.author)


class ModListFile:
    @staticmethod
    def read(path: Path) -> list[Mod]:
        try:
            with path.open("r") as f:
                text = f.read()
        except OSError as e:
            print(e)
            return None

        if re.search(r"^\s?[0-9]+\s?#.*$", text, re.MULTILINE):
            return [m for m in ModListV1Format.parse(text)]

        return [m for m in ModListV2Format.parse(text)]

    @staticmethod
    def write(path: Path, mods: MutableSequence, serializer: ModListSerializer) -> bool:
        try:
            with path.open("w+") as f:
                [f.write(line + "\n") for line in serializer.serialize(mods)]
        except OSError as e:
            print(e)
            return False
        return True



#!/usr/bin/env python3

from pathlib import Path
from typing import List, Optional

from . import util


class PathFinder:
    DEFAULT_GAME_PATHS = [
        ("~/GOG Games/RimWorld", "linux"),
        ("~/games/rimworld", "linux"),
        ("~/.local/share/Steam/steamapps/common/RimWorld", "linux"),
        ("/Applications/RimWorld.app/Mods", "darwin"),
        ("~/Library/Application Support/Steam/steamapps/common/RimWorld", "darwin"),
        ("C:/GOG Games/RimWorld/Mods", "win32"),
        ("C:/Program Files (x86)/Steam/steamapps/common/RimWorld", "win32"),
        ("C:/Program Files/Steam/steamapps/common/RimWorld", "win32"),
    ]

    DEFAULT_WORKSHOP_PATHS = [
        ("~/.local/share/Steam/steamapps/workshop/content/294100", "linux"),
        (
            "~/Library/Application Support/Steam/steamapps/workshop/content/294100",
            "darwin",
        ),
        (
            "C:/Program Files (x86)/Steam/steamapps/common/workshop/content/294100",
            "win32",
        ),
        ("C:/Program Files/Steam/steamapps/common/workshop/content/294100", "win32"),
    ]

    DEFAULT_CONFIG_PATHS = [
        ("~/Library/Application Support/RimWorld/", "darwin"),
        ("~/.config/unity3d/Ludeon Studios/RimWorld by Ludeon Studios", "linux"),
        ("~/AppData/LocalLow/Ludeon Studios/RimWorld by Ludeon Studios", "win32"),
    ]

    @staticmethod
    def _is_game_dir(p: Path) -> bool:
        if p.name == "Mods":
            for n in p.parent.iterdir():
                if n.name == "Version.txt":
                    return True
        return False

    @staticmethod
    def _is_workshop_dir(p: Path) -> bool:
        return (
            p.name == "294100"
            and p.parts[-2] == "content"
            and p.parts[-3] == "workshop"
        )

    @staticmethod
    def _is_config_dir(p: Path) -> bool:
        files_to_find = ["Config", "Saves"]
        child_names = [f.name for f in p.iterdir()]
        for target_name in files_to_find:
            if not target_name in child_names:
                return False
        return True

    @staticmethod
    def _search_root(p: Path, f) -> Optional[Path]:
        try:
            p = util.sanitize_path(p)
            for n in p.glob("**/"):
                if f(n):
                    return n
            return None
        except FileNotFoundError:
            return None

    @staticmethod
    def get_workshop_from_game_path(p: Path):
        p = util.sanitize_path(p)
        for index, dirname in enumerate(p.parts):
            if dirname == "steamapps":
                return Path(*list(p.parts[0:index])) / Path(
                    "steamapps/workshop/content/294100"
                )

    @staticmethod
    def _search_defaults(defaults: List[str], f) -> Optional[Path]:
        platform = util.platform()
        for path in [n[0] for n in defaults if n[1] == platform]:
            path = util.sanitize_path(path)
            if path := f(Path(path)):
                return path
        return None

    @classmethod
    def find_game(cls, p: Path) -> Optional[Path]:
        return cls._search_root(p, cls._is_game_dir)

    @classmethod
    def find_workshop(cls, p: Path) -> Optional[Path]:
        return cls._search_root(p, cls._is_workshop_dir)

    @classmethod
    def find_config(cls, p: Path) -> Optional[Path]:
        return cls._search_root(p, cls._is_config_dir)

    @classmethod
    def find_game_defaults(cls) -> Optional[Path]:
        return cls._search_defaults(cls.DEFAULT_GAME_PATHS, cls.find_game)

    @classmethod
    def find_workshop_defaults(cls) -> Optional[Path]:
        return cls._search_defaults(cls.DEFAULT_WORKSHOP_PATHS, cls.find_workshop)

    @classmethod
    def find_config_defaults(cls) -> Optional[Path]:
        return cls._search_defaults(cls.DEFAULT_CONFIG_PATHS, cls.find_config)
